from permpy.permset import PermSet
from permpy.permutation import Permutation
from permpy.pegpermutation import PegPermutation
from permpy.permclass import PermClass
from permpy.avclass import AvClass
from permpy.pegpermset import PegPermSet
from permpy.geometricgridclass import GeometricGridClass

# from permpy.InsertionEncoding import *
import permpy.RestrictedContainer

import doctest

doctest.testmod()



class Vector(tuple):
    def __new__(cls, v):
        for i in range(0, len(v)):
            v[i] = int(v[i])
            assert v[i] >= 0, "vector entries must be nonnegative integers"
        return tuple.__new__(cls, v)

    def meet(self, V):
        assert len(V) == len(self), "vectors must be same length to meet"
        new_vector = []
        for i in range(0, len(V)):
            new_vector.append(max(V[i], self[i]))
        return Vector(new_vector)

    def contained_in(self, V):
        assert len(V) == len(self), "vectors must be same length to check containment"
        for i in range(0, len(V)):
            if self[i] > V[i]:
                return False
        return True

    def norm(self):
        return sum(self)



from math import fabs
from sympy import *
from itertools import chain, combinations

from .permutation import Permutation
from .permset import *
from .permclass import *
from .vectorset import *


def powerset(iterable):
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(0, len(s) + 1))


class PegPermutation(Permutation):
    def __new__(cls, p, signs):
        if isinstance(p, int):
            p = list(str(p))
        assert len(p) == len(signs), "improper sign string"
        for i in range(0, len(signs)):
            assert (
                signs[i] == "+" or signs[i] == "-" or signs[i] == "."
            ), "improper sign string"

        return Permutation.__new__(cls, p)

    def __init__(self, p, signs):
        self.signs = list(signs)

    def __repr__(self):
        s = " "
        for i in range(0, len(self)):
            s += str(self[i] + 1) + self.signs[i] + " "
        return s

    def __hash__(self):
        return hash((tuple(self[:]), tuple(self.signs)))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def num_signs(self):
        return sum([1 for s in self.signs if s != "."])

    def sign_subset(self, P):
        for i in range(0, len(self)):
            t = (self.signs[i], P.signs[i])
            if t in [("+", "."), ("+", "-"), ("-", "+"), ("-", ".")]:
                return False
        return True

    def filling_vector(self):
        L = []
        for s in self.signs:
            if s == "+" or s == "-":
                L.append(2)
            else:
                L.append(1)
        return Vector(L)

    def all_dotted_monotone_intervals(self):
        mi = []
        difference = 0
        c_start = 0
        c_length = 0
        for i in range(0, len(self) - 1):
            if (
                self.signs[i] == "."
                and self.signs[i + 1] == "."
                and math.fabs(self[i] - self[i + 1]) == 1
                and (c_length == 0 or self[i] - self[i + 1] == difference)
            ):
                if c_length == 0:
                    c_start = i
                c_length += 1
                difference = self[i] - self[i + 1]
            else:
                if c_length != 0:
                    mi.append((c_start, c_start + c_length))
                c_start = 0
                c_length = 0
                difference = 0
        if c_length != 0:
            mi.append((c_start, c_start + c_length))
        return mi

    def is_compact(self):
        for i in range(0, len(self) - 1):
            if self[i] + 1 == self[i + 1]:
                s = "".join(self.signs[i : i + 2])
                if s == "++" or s == "+." or s == ".+":
                    return False
            elif self[i] - 1 == self[i + 1]:
                s = "".join(self.signs[i : i + 2])
                if s == "--" or s == "-." or s == ".-":
                    return False
        return True

    def is_compact_and_clean(self):
        if not self.is_compact():
            return False
        for i in range(0, len(self) - 1):
            if (
                self.signs[i] == "."
                and self.signs[i + 1] == "."
                and fabs(self[i] - self[i + 1]) == 1
            ):
                return False
        return True

    def clean_basis(self):
        dotted_intervals = self.all_dotted_monotone_intervals()
        in_intervals = []
        for (start, end) in dotted_intervals:
            in_intervals.extend(range(start, end + 1))
        for i in range(0, len(self)):
            if i not in in_intervals:
                dotted_intervals.append((i, i))
        dotted_intervals.sort()
        S = VectorSet()
        for (i, (start, end)) in enumerate(dotted_intervals):
            L = [0] * len(dotted_intervals)
            if start == end:
                if self.signs[start] == ".":
                    L[i] = 0
            else:
                length_to_avoid = end - start + 2
                L[i] = length_to_avoid
            for j in L:
                if j != 0:
                    if Vector(L).norm() > 0:
                        S.append(Vector(L))
                    break
        return S

    def split(self, indices):
        entries = list(self[:])
        signs = list(self.signs)
        indices.sort()
        offset = 0
        for index in indices:
            new_entries = entries[: index + offset]
            new_signs = signs[: index + 1 + offset]
            new_signs.append(signs[index + offset])

            if signs[index + offset] == "+":
                new_entries.append(entries[index + offset] - 0.5)
                new_entries.append(entries[index + offset] + 0.5)
            else:
                new_entries.append(entries[index + offset] + 0.5)
                new_entries.append(entries[index + offset] - 0.5)
            new_entries.extend(entries[index + offset + 1 :])
            new_signs.extend(signs[index + offset + 1 :])

            entries = Permutation.standardize(new_entries)
            signs = new_signs
            offset += 1

        return PegPermutation(entries, signs)

    def clean(self):
        if not self.is_compact():
            return PegPermutation(self, self.signs)
        dotted_intervals = self.all_dotted_monotone_intervals()
        copy = list(self[:])
        copysigns = list(self.signs[:])
        for (start, end) in dotted_intervals:
            type_of_interval = self[start + 1] - self[start]
            copysigns[start] = "+" if type_of_interval == 1 else "-"
            for i in range(start + 1, end + 1):
                copy[i] = -1
                copysigns[i] = -1
        while -1 in copy:
            copy.remove(-1)
            copysigns.remove(-1)
        return PegPermutation(copy, copysigns)

    def downset(self):
        return PermSet([self]).downset()

    def shrink_by_one(self):
        S = PermSet()

        for i in range(0, len(self)):
            if self.signs[i] != ".":
                new_signs = list(self.signs)
                new_signs[i] = "."

                S.add(PegPermutation(list(self), new_signs))

            else:
                if len(self) == 1:
                    continue
                new_signs = list(self.signs)
                new_signs.pop(i)

                new_entries = list(self)
                new_entries.pop(i)

                S.add(PegPermutation(new_entries, new_signs))

        return S

    def csgf(self, basis):
        x = Symbol("x")
        subsets = powerset(basis)
        f = 0
        for subset in subsets:
            V = VectorSet(subset)
            negpow = pow(-1, len(subset))
            fv = self.filling_vector()
            meet_B = V.meet_all() if len(V) != 0 else [1] * len(fv)
            xpow = fv.meet(meet_B).norm()
            f += negpow * pow(x, xpow) / pow(1 - x, self.num_signs())
        return f

    def reverse(self):
        entries = self[::-1]
        signs = self.signs[::-1]
        for i in range(0, len(signs)):
            if signs[i] == "-":
                signs[i] = "+"
            elif signs[i] == "+":
                signs[i] = "-"
        return PegPermutation(entries, signs)

    def involved_in(self, P):
        if not self.bounds_set:
            (self.lower_bound, self.upper_bound) = self.set_up_bounds()
            self.bounds_set = True
        L = list(self)
        n = len(L)
        p = len(P)
        # if n <= 1 and n <= p:
        # return True

        indices = [0] * n

        while indices[0] < p:
            if self.involvement_check(
                self.upper_bound, self.lower_bound, indices, P, 1
            ):
                if self.sign_subset(P):
                    return True
            indices[0] += 1

        return False

    def involvement_check(self, upper_bound, lower_bound, indices, q, next):
        if next == len(indices):
            return True
        lq = len(q)
        indices[next] = indices[next - 1] + 1
        while indices[next] < lq:
            if self.involvement_fits(
                upper_bound, lower_bound, indices, q, next
            ) and self.involvement_check(
                upper_bound, lower_bound, indices, q, next + 1
            ):
                if self.sign_subset(q):
                    return True
            indices[next] += 1
        return False

    def involvement_fits(self, upper_bound, lower_bound, indices, q, next):
        return (
            lower_bound[next] == -1 or q[indices[next]] > q[indices[lower_bound[next]]]
        ) and (
            upper_bound[next] == -1 or q[indices[next]] < q[indices[upper_bound[next]]]
        )



import fractions

from functools import reduce

from permpy.permutation import Permutation


def bend_list(perm):
    """Returns the list of indices at that the permutation changes
    direction. That is, the number of non-monotone consecutive triples of
    the permutation. A permutation p can be expressed as the concatenation
    of len(p.bend_list()) + 1 monotone segments."""
    raise NotImplementedError

    # this isn't quite correct....
    # return len(
    #     [
    #         i
    #         for i in range(1, len(perm) - 1)
    #         if (perm[i - 1] > perm[i] and perm[i + 1] > perm[i])
    #         or (perm[i - 1] < perm[i] and perm[i + 1] < perm[i])
    #     ]
    # )


def order(perm):
    L = map(len, perm.cycle_decomp())
    return reduce(lambda x, y: x * y // fractions.gcd(x, y), L)


def bonds(perm):
    numbonds = 0
    p = list(perm)
    for i in range(1, len(p)):
        if p[i] - p[i - 1] == 1 or p[i] - p[i - 1] == -1:
            numbonds += 1
    return numbonds


def majorindex(perm):
    sum = 0
    p = list(perm)
    n = perm.__len__()
    for i in range(0, n - 1):
        if p[i] > p[i + 1]:
            sum += i + 1
    return sum


def fixedptsplusbonds(perm):
    return perm.fixed_points() + perm.bonds()


def christiecycles(perm):
    # builds a permutation induced by the black and gray edges separately, and
    # counts the number of cycles in their product. used for transpositions
    p = list(perm)
    n = perm.__len__()
    q = [0] + [p[i] + 1 for i in range(n)]
    grayperm = range(1, n + 1) + [0]
    blackperm = [0 for _ in range(n + 1)]
    for i in range(n + 1):
        ind = q.index(i)
        blackperm[i] = q[ind - 1]
    newperm = []
    for i in range(n + 1):
        k = blackperm[i]
        j = grayperm[k]
        newperm.append(j)
    return Permutation(newperm).numcycles()


def othercycles(perm):
    # builds a permutation induced by the black and gray edges separately, and
    # counts the number of cycles in their product
    p = list(perm)
    n = perm.__len__()
    q = [0] + [p[i] + 1 for i in range(n)]
    grayperm = [n] + range(n)
    blackperm = [0 for _ in range(n + 1)]
    for i in range(n + 1):
        ind = q.index(i)
        blackperm[i] = q[ind - 1]
    newperm = []
    for i in range(n + 1):
        k = blackperm[i]
        j = grayperm[k]
        newperm.append(j)
    return Permutation(newperm).numcycles()


def sumcycles(perm):
    return perm.othercycles() + perm.christiecycles()


def maxcycles(perm):
    return max(perm.othercycles() - 1, perm.christiecycles())



from __future__ import print_function

from itertools import combinations_with_replacement as cwr


def pretty_out(pi, k, vert_line=True, by_lines=False, width=2):
    """Return a nice string to visualize `pi`.
    If `by_lines == True`, then will return the list of strings by line,
    in case you want to append some stuff to each line.

    """
    print(pi, k)
    lines = []
    n = len(pi)

    max_width = len(str(n + 1))  # This is the width of each value.
    if max_width > width:
        width = max_width

    blank = " " * width
    for val in range(n)[::-1]:
        idx = pi.index(val)
        line = blank * (idx) + str(val + 1).rjust(width) + blank * (n - idx - 1)
        lines.append(line)

    if vert_line:
        if k == 0:
            for idx in range(n):
                lines[idx] += " |"
        else:
            for idx in range(n):
                lines[idx] = lines[idx][: -width * k] + " |" + lines[idx][-width * k :]
    else:
        lines.insert(k, "-" * (width * n))

    if by_lines:
        return lines
    else:
        return "\n".join(lines)


def gen_compositions(n, k=0):
    """Generate all compositions (as lists) of `n` into `k` parts.
    If `k == 0`, then generate all compositions of `n`.

    """
    assert n >= k, f"Need weight to be at least length: {n} ≥ {k}"

    if k == 0:
        for i in range(1, n + 1):
            for c in gen_compositions(n, i):
                yield c
    else:
        if k == 1:
            yield [n]
        elif n == k:
            yield [1] * n
        else:
            for i in range(1, n - k + 2):
                for c in gen_compositions(n - i, k - 1):
                    yield c + [i]


def gen_weak_compositions(n, k):
    """Generate all weak compositions (as lists) of `n` into `k` parts."""
    for c in gen_compositions(n + k, k):
        yield [part - 1 for part in c]


def gen_interval_divisions(m, k, shift=0, reverse=False):
    """Generate all ways of splitting the interval `[1, m]` shifted up by `shift` into `k` pieces.

    Example:
        >>> list(gen_interval_divisions(4, 2))
        [
            [ ()          , (0, 1, 2, 3) ],
            [ (0,)        ,    (1, 2, 3) ],
            [ (0, 1)      ,       (2, 3) ],
            [ (0, 1, 2)   ,          (3,)],
            [ (0, 1, 2, 3),            ()]
        ]
    """
    if reverse:
        direction = -1
    else:
        direction = +1

    L = range(shift, shift + m)[::direction]
    for c in cwr(range(m + 1), k - 1):
        # For each choice of divisions...

        c = (0,) + c + (m,)
        yield [tuple(val for val in L[c[i] : c[i + 1]]) for i in range(k)]


def all_vertical_extensions(pi, m, k, verbose=False):
    """Given a permutation `pi`, generate all ways to add an increasing sequence
    of length `m` above its right `k` points.

    """
    n = len(pi)

    # Split pi on its last k elements.
    if k == 0:
        prefix = pi
        suffix = ()
    else:
        prefix = pi[:-k]
        suffix = pi[-k:]

    if verbose:
        print(f"Vertically extending (pi, m, k) = {(pi, m, k)}")
        print(f"prefix = {prefix}")
        print(f"suffix = {suffix}")

    for uppers in gen_interval_divisions(m, k + 1, shift=n):
        new_suffix = sum([uppers[i] + (suffix[i],) for i in range(k)], ()) + uppers[-1]

        if verbose:
            print(f"uppers = {uppers:20}, new_suffix = {new_suffix:20}")
            print(f"Yielding {prefix + new_suffix}.")

        yield prefix + new_suffix


def all_horizontal_extensions(pi, m, k, verbose=False):
    """Given a permutation `pi`, generate all ways to add an decreasing sequence
    of length `m` to the right of its upper `k` points.

    """

    tau = inverse(pi)
    n = len(tau)

    if k == 0:
        prefix = tau
        suffix = ()
    else:
        prefix = tau[:-k]
        suffix = tau[-k:]

    if verbose:
        print(f"Horizontally extending (pi, m, k) = {(pi,m,k)}")
        print(f"prefix = {prefix}")
        print(f"suffix = {suffix}")

    for uppers in gen_interval_divisions(m, k + 1, shift=n, reverse=True):
        new_suffix = sum([uppers[i] + (suffix[i],) for i in range(k)], ()) + uppers[-1]

        if verbose:
            print(f"uppers = {uppers:20}, new_suffix = {new_suffix:20}")
            print(f"Yielding the inverse of {prefix + new_suffix}.")

        yield inverse(prefix + new_suffix)


def inverse(pi):
    q = tuple(pi.index(val) for val in range(len(pi)))
    return q


def first_two_cells(n):
    """Return the set of initial configurations of points in the first two cells."""

    initial = ((), 0)
    R = set([initial])  # The set containing the empty tuple.

    S = set()

    for pi, k in R:
        for m in range(0, n + 1):
            S.update((tau, m) for tau in all_vertical_extensions(pi, m, k))

    T = set()

    for pi, k in S:
        if k == 0 and len(pi) != 0:
            T.add((pi, 0))
        else:
            for m in range(0, n - len(pi) + 1):
                T.update((tau, m) for tau in all_horizontal_extensions(pi, m, k))

    return T


def add_two_cells(R, n):
    S = set()
    for pi, k in R:
        S.add((pi, 0))
        for m in range(1, n - len(pi) + 1):
            S.update((tau, m) for tau in all_vertical_extensions(pi, m, k))

    T = set()
    for pi, k in S:
        T.add((pi, 0))
        for m in range(1, n - len(pi) + 1):
            T.update((tau, m) for tau in all_horizontal_extensions(pi, m, k))

    return T


if __name__ == "__main__":
    pass

    # for pi in all_vertical_extensions((0, 1), 1, 0):
    #     print(pretty_out(pi, 1))
    #     print("-" * 6)



""" This is working code that may or may not be added to a class at some point."""

from __future__ import print_function
from permpy import *
from math import factorial
import itertools

from permpy.RestrictedContainer import *


def expected_basis(B):
    return PermSet([expected_basis_element(P) for P in B]).minimal_elements()


def expected_basis_element(P):
    if P[0] == len(P) - 1:
        return P
    else:
        return Permutation([len(P) + 2] + list(P))


def powerset(iterable):
    s = list(iterable)
    return itertools.chain.from_iterable(
        itertools.combinations(s, r) for r in range(len(s) + 1)
    )


def check_combos(max_length, size_of_combos=0, verbose=False):
    bad_ones = 0
    so_far = 0
    permlist = []
    for i in range(2, max_length + 1):
        permlist.extend(PermSet.all(i))
    if size_of_combos == 0:
        to_check = powerset(permlist)
    else:
        to_check = itertools.combinations(permlist, size_of_combos)
    for basis in to_check:
        if len(basis) < 1:
            continue
        so_far += 1
        print("----------------------------------------------------")
        expected = expected_basis(basis)
        B = make_restricted_container_class(basis).guess_basis()
        if B == expected:
            print(f"{basis}  -->  {B}  :  ok")
        else:
            bad_ones += 1
            print(f"{basis}  -->  {B}  : WHOA!")
        if verbose and so_far % 10 == 0:
            print(f"----------------------------------------------------")
            print(f"----------------------------------------------------")
            print(f" === SO FAR: {bad_ones} BAD BASES OUT OF {so_far} === ")
            print(f"----------------------------------------------------")
    print(f"----------------------------------------------------")
    print(f"----------------------------------------------------")
    print(f"----------------------------------------------------")
    print(f"=== THERE WERE {bad_ones} BAD BASES OUT OF {so_far} ===")
    print(f"----------------------------------------------------")


def check_singleton_restricted_containers():
    for length in range(2, 6):
        print("----------------------------------------------------")
        for index in range(0, factorial(length)):
            print("----------------------------------------------------")
            P = Permutation(index, length)
            expected = expected_basis(P)
            B = make_restricted_container_class([P]).guess_basis()
            if B == expected:
                print(f"{P}  -->  {B}  :  ok")
            else:
                print(f"{P}  -->  {B}  : WHOA!")
    print("----------------------------------------------------")


def make_restricted_container_class(basis, length=7):
    basis = [Permutation(b) for b in basis]
    D = [PermSet([])]
    for i in range(1, length + 1):
        A = RestrictedContainer(basis, i)
        D.append(generate_from_restricted_container(A))
    return PermClass(D)


def generate_from_restricted_container(container):
    perms_generated = PermSet()
    queue = [container]
    while len(queue) > 0:
        machine = queue.pop(0)
        container_size = machine.container_size()
        input_size = machine.input_size()
        if container_size == 0 and input_size == 0:
            output = machine.output()
            output_permutation = Permutation(output)
            perms_generated.add(output_permutation)
        else:
            if container_size > 0:
                queue.append(machine.pop())
            for open_spot in range(0, container_size + 1):
                new_machine = machine.push(open_spot)
                if new_machine != -1:
                    queue.append(new_machine)
    return perms_generated


def color_4231(p):
    red_entries = []
    blue_entries = []

    for entry in p:
        if Permutation(blue_entries + [entry]).involves(Permutation(312)):
            red_entries.append(entry)
        else:
            blue_entries.append(entry)

    return (red_entries, blue_entries)


def label_4231(p, new_version=False):
    (red_entries, blue_entries) = color_4231(p)
    word = ""
    for (index, entry) in enumerate(p):
        if new_version and index in p.ltrmax():
            word += "D"
        elif new_version and index in p.rtlmin():
            word += "A"
        elif entry in red_entries:
            word += "A" if index in p.rtlmin() else "B"
        else:
            word += "D" if index in p.ltrmax() else "C"
    return (word, "".join([word[i] for i in p.inverse()]))


def color_1324(p):
    red_entries = []
    blue_entries = []

    for entry in p:
        if (len(blue_entries) > 0 and entry > min(blue_entries)) or Permutation(
            red_entries + [entry]
        ).involves(Permutation(132)):
            blue_entries.append(entry)
        else:
            red_entries.append(entry)

    return (red_entries, blue_entries)


def label_1324(p, new_version=False):
    (red_entries, blue_entries) = color_1324(p)
    word = ""
    for (index, entry) in enumerate(p):
        if new_version and index in p.rtlmax():
            word += "D"
        elif entry in red_entries:
            word += "A" if index in p.ltrmin() else "B"
        else:
            word += "D" if index in p.rtlmax() else "C"
    return (word, "".join([word[i] for i in p.inverse()]))


def check_pattern(word_pairs, pattern):
    n = [0, 0]
    for (x, y) in word_pairs:
        if x.find(pattern) != -1:
            n[0] += 1
        if y.find(pattern) != -1:
            n[1] += 1
    if n[0] == 0 or n[1] == 0:
        print(f"\n\n{'!'*40}\n\t\t{pattern}: ({str(n[0])}, {str(n[1])})\n\n")
    return n


def check_pattern_list(word_pairs, patterns):
    d = dict()
    for pattern in patterns:
        d[pattern] = check_pattern(word_pairs, pattern)
    return d


def allstrings(alphabet, length):
    c = []
    for i in range(length):
        c = [[x] + y for x in alphabet for y in c or [[]]]
    return ["".join(a) for a in c]


def nc_contains(w, u):
    seen = 0
    for i in range(0, len(w)):
        if w[i] == u[seen]:
            seen += 1
            if seen == len(u):
                return True
    return False


def check_nc_pattern(word_pairs, pattern):
    n = [0, 0]
    for (x, y) in word_pairs:
        if nc_contains(x, pattern):
            n[0] += 1
        if nc_contains(y, pattern):
            n[1] += 1
    return n


def check_nc_pattern_list(word_pairs, patterns):
    d = dict()
    for pattern in patterns:
        d[pattern] = check_nc_pattern(word_pairs, pattern)
    return d


# implements rsk:
class Tab(object):
    def __init__(self, P, Q):
        self.P = P
        self.Q = Q

    def __repr__(self):
        l = len(self.P[0])
        h = len(self.P)
        sP = "P = |"
        sQ = "Q = |"
        for i in range(h):
            for j in range(len(self.P[i])):
                sP += ("%2i" % self.P[i][j]) + "|"
                sQ += ("%2i" % self.Q[i][j]) + "|"
            sP += "\n    |"
            sQ += "\n    |"
        return sP[:-6] + "\n" + sQ[:-6]


def rsk(perm):
    p = [i + 1 for i in perm[::-1]]
    q = 1
    P = [[]]
    Q = [[]]

    def insert(k, r, q1):
        if len(P) < r + 1:
            P.append([])
            Q.append([])
        if not P[r] or k > P[r][-1]:
            P[r].append(k)
            Q[r].append(q1)
        else:
            n = len(P[r])
            i = n - 1
            while True:
                b = P[r][i]
                if k < b:
                    if i > 0:
                        i -= 1
                    else:
                        P[r][0] = k
                        break
                if k > b:
                    b = P[r][i + 1]
                    P[r][i + 1] = k
                    break
            insert(b, r + 1, q)

    for i in range(len(perm) - 1, -1, -1):
        insert(p.pop(), 0, q)
        q += 1
    return Tab(P, Q)


def rsk_shape(perm):
    P = rsk(perm).P
    return [len(P[i]) for i in range(0, len(P))]


def shape_contains(A, B):
    return len(A) >= len(B) and all(b <= a for a, b in zip(A, B))


def check_stat(f, L1, L2, l=8):
    C = [sorted([f(P) for P in T[l]]) for T in L1]
    D = [sorted([f(P) for P in T[l]]) for T in L2]
    return [(i, j) for i in range(len(L1)) for j in range(len(L2)) if C[i] == D[j]]


def check_stat_list(fs, L1, L2, l=8):
    C = [sorted([tuple([f(P) for f in fs]) for P in T[l]]) for T in L1]
    D = [sorted([tuple([f(P) for f in fs]) for P in T[l]]) for T in L2]
    # done = []
    # vals = []
    # for thing in C[0]:
    # if thing in done:
    # continue
    # done.append(thing)
    # vals.append(C[0].count(thing))
    # print C[0].count(thing),'\t->\t',thing
    # print max(vals)
    return [(i, j) for i in range(len(L1)) for j in range(len(L2)) if C[i] == D[j]]


def size_value_blocks(L):
    if len(L) == 0:
        return []
    L = sorted(list(L))
    T = []
    s = 0
    for i in range(len(L) - 1):
        s += 1
        if L[i] + 1 not in L:
            T.append(s)
            s = 0
    s += 1
    T.append(s)
    return sorted(T)


def rhs_blocks(P):
    right_side = P[P.index(0) :]
    rhs_sorted = sorted(right_side)
    sorted_lookup = sorted(enumerate(right_side), key=lambda i: i[1])
    rhs_blocks = []
    b_sizes = []
    this_block_min = 0
    this_block_size = 1
    for i in range(len(rhs_sorted) - 1):
        if rhs_sorted[i] + 1 in rhs_sorted:
            this_block_size += 1
        else:
            rhs_blocks.append(
                [
                    t[1]
                    for t in sorted(
                        [
                            sorted_lookup[i]
                            for i in range(
                                this_block_min, this_block_min + this_block_size
                            )
                        ],
                        key=lambda i: i[0],
                    )
                ]
            )
            this_block_min = i + 1
            b_sizes.append(this_block_size)
            this_block_size = 1
    b_sizes.append(this_block_size)
    # if this_block_size == 1:

    rhs_blocks.append(
        [
            t[1]
            for t in sorted(
                [
                    sorted_lookup[i]
                    for i in range(this_block_min, this_block_min + this_block_size)
                ],
                key=lambda i: i[0],
            )
        ]
    )
    return rhs_blocks


def size_LR_max_contig_blocks(P):
    rhs_bs = rhs_blocks(P)
    sizes = []
    for i in range(1, len(rhs_bs)):
        b = rhs_bs[i]
        P = Perm(b)
        vs = [P[i] for i in P.ltrmax()]
        v = max(vs)
        j = 1
        while v - j in vs:
            j += 1
        sizes.append(j)
    return sizes


def num_value_blocks(L):
    if len(L) == 0:
        return 0
    L = sorted(list(L))
    n = 1
    for i in range(len(L) - 1):
        if L[i] + 1 not in L:
            n += 1
    return n


def longest_init_asc(L):
    n = 1
    for i in range(1, len(L)):
        if L[i] > L[i - 1]:
            n += 1
        else:
            return n
    return n


def longest_init_desc(L):
    return longest_init_asc(Perm(L).complement())


def split_and_check_perm(P):
    (a, b) = (P[: P.index(0)], P[P.index(0) :])
    return (num_value_blocks(a), num_value_blocks(b))


def split_size_value_blocks(P):
    (a, b) = (P[: P.index(0)], P[P.index(0) :])
    return (size_value_blocks(a), size_value_blocks(b))


def up_jump_lengths(P):
    L = []
    for i in range(len(P) - 1):
        if P[i + 1] - P[i] > 0:
            L.append(P[i + 1] - P[i])
        else:
            L.append(0)
    return L


def check_stats(L1, L2, l=8, add_syms=False, tups=1):
    """Not sure what this does.

    ME: I've updated some of this, but I don't know what many of these even do.
    """
    if add_syms:
        L1 = [
            item
            for sublist in [list(PermSet(l1).all_syms()) for l1 in L1]
            for item in sublist
        ]
        L2 = [
            item
            for sublist in [list(PermSet(l2).all_syms()) for l2 in L2]
            for item in sublist
        ]
    stats = [
        ("number of descents", Perm.num_descents),
        ("position of descents", Perm.descents),
        ("number of ascents", Perm.num_ascents),
        ("position of ascents", Perm.ascents),
        ("number of ltrmin", Perm.num_ltr_min),
        ("positions of ltrmin", Perm.ltr_min),
        ("values of ltrmin", lambda P: [P[i] for i in P.ltr_min()]),
        ("number of ltrmax", Perm.num_ltr_max),
        ("positions of ltrmax", Perm.ltr_max),
        ("values of ltrmax", lambda P: [P[i] for i in P.ltr_max()]),
        ("number of rtlmin", Perm.num_rtl_min),
        ("positions of rtlmin", Perm.rtl_min),
        ("values of rtlmin", lambda P: [P[i] for i in P.rtl_min()]),
        ("number of rtlmax", Perm.num_rtl_max),
        ("positions of rtlmax", Perm.rtl_max),
        ("values of rtlmax", lambda P: [P[i] for i in P.rtl_max()]),
        ("is sum decomposable", Perm.sum_decomposable),
        ("number of sum components", lambda P: len(P.sum_decomposition())),
        ("is skew decomposable", Perm.skew_decomposable),
        ("number of skew components", lambda P: len(P.skew_decomposition())),
        (
            "number of inflations of simples",
            lambda P: P.skew_decomposable() or P.sum_decomposable(),
        ),
        ("length of simple quotient", lambda P: len(P.decomposition()[0])),
        ("number of simple permutations", Perm.is_simple),
        ("number of bonds", Perm.bonds),
        (
            "+ bonds",
            lambda P: len([i for i in range(len(P) - 1) if P[i + 1] == P[i] + 1]),
        ),
        (
            "- bonds",
            lambda P: len([i for i in range(len(P) - 1) if P[i + 1] == P[i] - 1]),
        ),
        (
            "position of + bonds",
            lambda P: [i for i in range(len(P) - 1) if P[i] + 1 == P[i + 1]],
        ),
        (
            "position of - bonds",
            lambda P: [i for i in range(len(P) - 1) if P[i] - 1 == P[i + 1]],
        ),
        ("number of bends", Perm.bends),
        (
            "position of bends",
            lambda P: [
                i
                for i in range(1, len(P) - 1)
                if (P[i - 1] > P[i] and P[i + 1] > P[i])
                or (P[i - 1] < P[i] and P[i + 1] < P[i])
            ],
        ),
        (
            "number of valleys",
            lambda P: len(
                [i for i in range(1, len(P) - 1) if P[i - 1] > P[i] and P[i + 1] > P[i]]
            ),
        ),
        (
            "position of valleys",
            lambda P: [
                i for i in range(1, len(P) - 1) if P[i - 1] > P[i] and P[i + 1] > P[i]
            ],
        ),
        (
            "number of peaks",
            lambda P: len(
                [i for i in range(1, len(P) - 1) if P[i - 1] < P[i] and P[i + 1] < P[i]]
            ),
        ),
        (
            "position of peaks",
            lambda P: [
                i for i in range(1, len(P) - 1) if P[i - 1] < P[i] and P[i + 1] < P[i]
            ],
        ),
        (
            "number of exceedances",
            lambda P: len([i for i in range(len(P)) if P[i] > i]),
        ),
        ("position of exceedances", lambda P: [i for i in range(len(P)) if P[i] > i]),
        ("value of first entry", lambda P: P[0]),
        ("value of second entry", lambda P: P[1]),
        ("value of second to last entry", lambda P: P[len(P) - 2]),
        ("value of last entry", lambda P: P[len(P) - 1]),
        ("position of 1", lambda P: [i for i in range(len(P)) if P[i] == 0]),
        ("position of 2", lambda P: [i for i in range(len(P)) if P[i] == 1]),
        ("position of n-1", lambda P: [i for i in range(len(P)) if P[i] == len(P) - 2]),
        ("position of n", lambda P: [i for i in range(len(P)) if P[i] == len(P) - 1]),
        ("major index", Perm.majorindex),
        ("longest ascending run", Perm.longestrunA),
        ("longest descending run", Perm.longestrunD),
        ("longest run", Perm.longestrun),
        ("number of inversions", Perm.inversions),
        (
            "number of top steps",
            lambda P: len(
                [
                    i
                    for i in [i for i in range(len(P) - 1) if P[i] > P[i + 1]]
                    if i in P.rtlmax() and i + 1 in P.rtlmax() and P[i] - 1 == P[i + 1]
                ]
            ),
        ),
        (
            "position of top steps",
            lambda P: [
                i
                for i in [i for i in range(len(P) - 1) if P[i] > P[i + 1]]
                if i in P.rtlmax() and i + 1 in P.rtlmax() and P[i] - 1 == P[i + 1]
            ],
        ),
        (
            "number of fixed points",
            lambda P: len([i for i in range(len(P)) if P[i] == i]),
        ),
        ("position of fixed points", lambda P: [i for i in range(len(P)) if P[i] == i]),
        ("number of disjoint cycles", Perm.num_disjoint_cycles),
        ("number of involutions", Perm.is_involution),
        ("rank encodings", Perm.rank_encoding),
        ("max rank", lambda P: max(P.rank_encoding())),
        # ('num rtlmax / ltrmin layers', Perm.num_rtlmax_ltrmin_layers),
        # ('num ltrmax / rtlmin layers', lambda P : P.reverse().num_rtlmax_ltrmin_layers()),
        # ('num rtlmax / ltrmin layers contracted', lambda P : P.contract_inc_bonds().num_rtlmax_ltrmin_layers()),
        ("occurrences of 123", lambda P: P.occurrences(Perm(123))),
        ("occurrences of 132", lambda P: P.occurrences(Perm(132))),
        ("occurrences of 213", lambda P: P.occurrences(Perm(213))),
        ("occurrences of 231", lambda P: P.occurrences(Perm(231))),
        ("occurrences of 312", lambda P: P.occurrences(Perm(312))),
        ("occurrences of 321", lambda P: P.occurrences(Perm(321))),
        ("occurrences of 2413", lambda P: P.occurrences(Perm(2413))),
        # ('split and count value-blocks', split_and_check_perm),
        # ('size_LR_max_contig_blocks', size_LR_max_contig_blocks),
    ]

    C1 = [Av(B, l) for B in L1]
    C2 = [Av(B, l) for B in L2]

    print("")
    if tups == 1:
        for (name, f) in stats:
            equivs = check_stat(f, C1, C2, l)
            print(name, ":")
            if len(equivs) == 0:
                print("\tnone\n")
                continue
            for (x, y) in equivs:
                print("\t", L1[x], "~", L2[y])
            print("")
    else:
        fs = itertools.combinations(stats, tups)
        for stat_set in fs:
            equivs = check_stat_list(
                [stat_set[i][1] for i in range(len(stat_set))], C1, C2, l
            )
            print("")
            print(
                "[", (" / ".join([stat_set[i][0] for i in range(len(stat_set))])), "]"
            )
            if len(equivs) == 0:
                print("\tnone\n")
                continue
            for (x, y) in equivs:
                print("\t", L1[x], "~", L2[y])


def kill_syms(bases):
    if len(bases) == 0:
        return bases
    bases = list(bases)
    new_bases = [PermSet(bases[0])]

    for B in bases[1:]:
        syms = [PermSet(K) for K in PermSet(B).all_syms()]
        if not any([BSym in new_bases for BSym in syms]):
            new_bases.append(PermSet(B))
    return new_bases


def kill_auts(basis_so_far, additions):
    auts = [
        lambda P: P.reverse(),
        lambda P: P.complement(),
        lambda P: P.reverse().complement(),
        lambda P: P.inverse(),
        lambda P: P.inverse().reverse(),
        lambda P: P.inverse().complement(),
        lambda P: P.inverse().reverse().complement(),
    ]

    good_auts = [a for a in auts if all([P == a(P) for P in basis_so_far])]

    keep_additions = []
    for beta in additions:
        if not any([a(beta) in keep_additions for a in good_auts]):
            keep_additions.append(beta)

    return keep_additions


def hunt_for_enumeration(enum, max_len=8, check_ahead=2, verbose=False):
    bases = [[]]
    done_bases = []
    # on_length = 1

    while len(bases) > 0:
        B = bases[0]
        Bshow = list(B)
        Bshow.sort()
        Bshow.sort(key=len)
        bases = bases[1:]
        print("Evaluating basis: ", list(Bshow), "\t(", len(bases), "more in queue )")
        C = AvClass(B, 6)
        diffs = [i for i in range(1, len(C)) if len(C[i]) != enum[i - 1]]
        if len(diffs) == 0:
            for i in range(7, max_len + 1):
                C.extend_to_length(i)
                diffs = [i for i in range(1, len(C)) if len(C[i]) != enum[i - 1]]
                if len(diffs) > 0:
                    break
        if len(diffs) == 0:
            # Now check all the way
            C.extend_to_length(max_len + check_ahead)
            if all([len(C[i]) == enum[i - 1] for i in range(1, len(C))]):
                print("\tPerfect match!")
                done_bases.append(B)
        else:
            first_diff = min(diffs)
            if len(C[first_diff]) <= enum[first_diff - 1]:
                if verbose:
                    print("\tNo chance.")
                continue
            if verbose:
                print("\tCurrent Enum:", [len(C[i]) for i in range(1, len(C))])
            if len(B) == 0:
                to_try = sorted([P for P in C[first_diff] if P.is_representative()])
            elif max([len(b) for b in B]) < first_diff:
                # try all basis elements in lex order
                to_try = sorted(list(C[first_diff]))
                if verbose:
                    print("\tPossible new basis elts:", to_try)
            else:
                # try all basis elements after last one (lex) of cur length in B
                to_try = sorted(list(C[first_diff]))
                to_try = [
                    b for b in to_try if b > max([t for t in B if len(t) == first_diff])
                ]
                if verbose:
                    print("\tPossible new basis elts:", to_try)
            if len(to_try) == 0:
                # nothing else to try of this length
                continue
            to_try = kill_auts(B, to_try)
            k = len(to_try)
            ci = 0
            for new_b in to_try:
                ci += 1
                new_B = PermSet(B).union(PermSet([new_b]))
                if verbose:
                    print(
                        "\t\tTrying basis:", sorted(list(new_B)), "\t(", ci, "/", k, ")"
                    )
                elif ci % 25 == 0:
                    print("\t(", ci, "/", k, ")")
                new_C = AvClass(new_B, first_diff + check_ahead)
                if all(
                    [
                        len(new_C[i]) >= enum[i - 1]
                        for i in range(first_diff + 1, first_diff + check_ahead + 1)
                    ]
                ):
                    if verbose:
                        print(
                            "\t\t\tOkay. New enum:",
                            [len(new_C[i]) for i in range(1, len(new_C))],
                        )

                        # print("\t\t\t\tMatches check_ahead: checking all the way.")
                        # new_C.extend_to_length(max_len+check_ahead)
                        # if all([len(new_C[i]) == enum[i-1] for i in range(1,max_len+check_ahead+1)]):
                        #   print("\t\t\t\t\tMatches all the way!")
                        #   done_bases.append(new_B)
                        #   continue
                        # else:
                        #   print("\t\t\t\tNot yet a perfect match.")
                    bases.append(new_B)
                else:
                    if verbose:
                        print("\t\t\tToo Small.")
        # Kill syms in list
        old_size = len(bases)
        bases = kill_syms(bases)
        new_size = len(bases)
        if verbose:
            print("Killed off symmetries: ", old_size, "->", new_size)

    return [list(B) for B in done_bases]


# def all_rtlmax_end(P,bound=1):
#   L = P.rtlmax()
#   return L == range(len(P)-len(L), len(P)) and len(L) >= bound

# from collections import defaultdict
# import csv

# ltd = 10

# def prof(P):
#   return tuple(tupify([f[1](P) for f in stats]))

# def tupify(L):
#   S = []
#   for p in L:
#     if isinstance(p,list):
#       S.append(tuple(p))
#     else:
#       S.append(p)
#   return tuple(S)

# def sf(P):
#   return str(P).replace(' ', '')

# def pos_of_least_val_block_inc(P):
#   L = [P.index(0)]
#   v = 1
#   while P.index(v) > P.index(v-1) and v < len(P):
#     L.append(P.index(v))
#     v += 1
#   return L

# def pos_of_greatest_val_block_inc(P):
#   L = [P.index(len(P)-1)]
#   v = len(P)-1
#   while P.index(v) > P.index(v-1):
#     L.append(P.index(v-1))
#     v -= 1
#   return L[::-1]


# stats = [
#   ('position of descents', lambda P : [i for i in range(len(P)-1) if P[i] > P[i+1]]),
#   ('positions of ltrmin', Perm.ltrmin),
#   ('positions of rtlmax', Perm.rtlmax),
#   ('position of + bonds', lambda P : [i for i in range(len(P)-1) if P[i]+1==P[i+1]]),
#   # ('least_val_block', pos_of_least_val_block_inc),
#   # ('greatest_val_block', pos_of_greatest_val_block_inc),
# ]

# av_2143 = defaultdict(set)
# av_3142 = defaultdict(set)

# C = Av([3142], ltd)
# D = Av([2143,415263], ltd)

# for P in C[ltd]:
#   if not P.skew_decomposable() and P not in D[ltd] and all_rtlmax_end(P,2) and P.num_inc_bonds() == 0:
#     av_3142[prof(P)].add(P)

# for P in D[ltd]:
#   if not P.skew_decomposable() and P not in C[ltd] and all_rtlmax_end(P,2) and P.num_inc_bonds() == 0:
#     av_2143[prof(P)].add(P)

# S = set(av_2143.keys()+av_3142.keys())

# with open('data.csv', 'wb') as cf:
#   w = csv.writer(cf)
#   for k in S:
#     w.writerow(list(k)+[' / '.join([sf(P) for P in av_2143[k]])]+[' / '.join([sf(P) for P in av_3142[k]])])



from collections import Counter, namedtuple

# X = namedtuple('X', ['a', 'b', 'c'])
X = namedtuple("X", ["v", "s"])


def next_states(S, mul):
    """Compute the states that would arise from the given state."""
    to_return = Counter()

    def include(new_v, new_s):
        to_return[X(new_v, new_s)] += mul

    if len(S.v) == 4:
        a, b, c, d = S.v

        if S.s == 1:
            include((b, c, d, 1), 2)

            for k in range(1, c + d):
                include((k, -1, 0), 1)

            include((a, b, c + 1, d), 1)

            for k in range(1, b):
                include((k, -1, 0), 1)

            for k in range(0, a + 1):
                include((k, -1, c + d), 1)

        elif S.s == 2:

            include((b + c, d, 1), 0)

            for k in range(1, d):
                include((k, -1, 0), 1)

            include((a, b, c, d + 1), 2)

            for k in range(1, b + c):
                include((k, -1, d), 2)

            for k in range(0, a + 1):
                include((k, -1, d), 1)
    elif len(S.v) == 3:

        a, b, c = S.v

        if S.s == 0:
            if c > 0:
                include((b, c, 1), 0)

            if b > 0:
                for k in range(1, c):
                    include((b, k, 1, c - k), 1)
            else:
                for k in range(1, c):
                    include((k, 1, c - k), 3)

            include((a, b, c + 1), 0)

            for k in range(1, b):
                include((k, -1, c), 2)

            if b > 0:
                if a > 0:
                    for k in range(0, a + 1):
                        include((k, -1, c), 1)
                else:
                    include((0, -1, c), 2)

        elif S.s == 1:
            include((c, -1, 0), 1)

            for k in range(1, c):
                include((k, -1, 0), 1)

            include((a, -1, c + 1), 1)

            for k in range(0, a + 1):
                include((k, -1, c), 1)

        elif S.s == 2:

            include((a, -1, 1), 1)

            for k in range(1, c):
                include((k, -1, 0), 1)

            include((a, -1, c + 1), 1)

            for k in range(0, a + 1):
                include((k, -1, c), 2)
        elif S.s == 3:
            assert a > 0 and b > 0 and c > 0

            include((a, b, c, 1), 2)

            for k in range(1, b + c):
                include((k, -1, 0), 1)

            include((a, b + 1, c), 3)

            for k in range(1, a):
                include((k, -1, 0), 1)

            include((0, -1, b + c), 1)
        else:
            raise Exception

    else:
        raise Exception

    return to_return


def iterate(states):
    """Iterate the current Counter of states once."""
    ns = sum([next_states(S, mul) for S, mul in states.items()], Counter())
    return ns


def states(n):
    """Return the states corresponding to permutations of length n in NS."""
    states = Counter()
    states[X(0, 0, 0)] = 1

    for _ in range(n):
        states = iterate(states)

    return states


def pstring(state, num):
    """Return a nice string give a state and its count.

    Example:
    >>> pstring(X(0,-1,1),4)
    ( 0, *, 1) :  4

    """
    if len(state.v) == 4:
        a, b, c, d = state.v

        return f"({a:2},{b:2},{c:2},{d:2})_{state.s} : {num:2}"
    elif len(state.v) == 3:
        a, b, c = state.v
        if b == -1:
            b = " *"
        return f"({a:2},{b:2},{c:2})_{state.s}    : {num:2}"


if __name__ == "__main__":
    S = Counter([X((0, 0, 0), 0)])

    N = 6

    for n in range(1, N + 1):
        print(f"\nn = {n-1} -> {n}\n")

        for state, num in sorted(S.items()):
            NS = next_states(state, num)
            if 0 < sum(NS.values()) // num < 100:
                print(f"{pstring(state, num)} -> ")
                for nstate, nnum in sorted(NS.items()):
                    print("\t" + pstring(nstate, nnum // num))
            # print("\t" + pstring(state, num))

        S = iterate(S)

        print(f"{n:2}, {sum(S.values()):10}")



import itertools

from math import gcd


def lcm(L):
    result = 1
    for val in L:
        result *= val // gcd(result, val)
    return result


class PermutationMiscMixin:
    """Contains various functions for Permutation to inherit."""

    @classmethod
    def one_cycles(cls, n):
        """Generate those permutations of length n that consist of one cycle."""
        for pi in itertools.permutations(range(n - 1)):
            cycle = [n - 1] + list(pi)
            tau = [None for _ in range(n)]
            for idx, val in enumerate(cycle[:-1]):
                tau[val] = cycle[idx + 1]
            tau[cycle[-1]] = cycle[0]
            yield (Permutation(tau), cycle)

    def cycle_decomp(self):
        """Return the cycle decomposition of the permutation.
        Return as a list of cycles, each of which is represented as a list.

        Examples:
                >>> Permutation(53814276).cycle_decomp()
                [[4, 3, 0], [6], [7, 5, 1, 2]]

        """
        not_seen = set(self)
        cycles = []
        while len(not_seen) > 0:
            a = max(not_seen)
            cyc = [a]
            not_seen.remove(a)
            b = self[a]
            while b in not_seen:
                not_seen.remove(b)
                cyc.append(b)
                b = self[b]
            cycles.append(cyc)
        return cycles[::-1]

    def cycles(self):
        """Return the cycle notation representation of the permutation."""
        stringlist = [
            "( " + " ".join([str(x + 1) for x in cyc]) + " )"
            for cyc in self.cycle_decomp()
        ]
        return " ".join(stringlist)

    def order(self):
        """Return the grou-theotric order of self."""
        L = map(len, self.cycle_decomp())
        return lcm(L)

    def children(self):
        """Return all patterns contained in self of length one less than the permutation."""
        return self.covers()

    def shrink_by_one(self):
        """Return all patterns contained in self of length one less than the permutation."""
        return self.covers()

    def all_right_extensions(self, max_length, l, S):
        if l == max_length:
            return S
        else:
            re = self.right_extensions()
        for p in re:
            S.add(p)
            S = p.all_right_extensions(max_length, l + 1, S)
        return S

    def sum_indecomposable_sequence(self):
        S = self.downset()
        return [
            len([p for p in S if len(p) == i and not p.sum_decomposable()])
            for i in range(1, max([len(p) for p in S]) + 1)
        ]

    # def sum_indec_bdd_by(self, n):
    #     l = [1]
    #     S = list(self.children())
    #     while len(S) > 0 and len(S[0]) > 0:
    #         l = [len([s for s in S if not s.sum_decomposable()])] + l
    #         if l[0] > n:
    #             return False
    #         S = list(permset.PermSet(S).layer_down())
    #     return True

    def contains_locations(self, Q):
        locs = []
        sublocs = itertools.combinations(range(len(self)), len(Q))
        for subloc in sublocs:
            if Permutation([self[i] for i in subloc]) == Q:
                locs.append(subloc)

        return locs

    def rank_val(self, i):
        return len([j for j in range(i + 1, len(self)) if self[j] < self[i]])

    def rank_encoding(self):
        return [self.rank_val(i) for i in range(len(self))]

    def num_rtlmax_ltrmin_layers(self):
        return len(self.rtlmax_ltrmin_decomposition())

    def rtlmax_ltrmin_decomposition(self):
        P = Permutation(self)
        num_layers = 0
        layers = []
        while len(P) > 0:
            num_layers += 1
            positions = sorted(list(set(P.rtlmax() + P.ltrmin())))
            layers.append(positions)
            P = Permutation([P[i] for i in range(len(P)) if i not in positions])
        return layers

    def contains_locations(self, Q):
        locs = []
        sublocs = itertools.combinations(range(len(self)), len(Q))
        for subloc in sublocs:
            if Permutation([self[i] for i in subloc]) == Q:
                locs.append(subloc)

        return locs

    def rank_val(self, i):
        return len([j for j in range(i + 1, len(self)) if self[j] < self[i]])

    def rank_encoding(self):
        return [self.rank_val(i) for i in range(len(self))]

    def num_rtlmax_ltrmin_layers(self):
        return len(self.rtlmax_ltrmin_decomposition())

    def rtlmax_ltrmin_decomposition(self):
        P = Permutation(self)
        num_layers = 0
        layers = []
        while len(P) > 0:
            num_layers += 1
            positions = sorted(list(set(P.rtlmax() + P.ltrmin())))
            layers.append(positions)
            P = Permutation([P[i] for i in range(len(P)) if i not in positions])
        return layers

    def num_inc_bonds(self):
        return len([i for i in range(len(self) - 1) if self[i + 1] == self[i] + 1])

    def num_dec_bonds(self):
        return len([i for i in range(len(self) - 1) if self[i + 1] == self[i] - 1])

    def num_bonds(self):
        return len(
            [
                i
                for i in range(len(self) - 1)
                if self[i + 1] == self[i] + 1 or self[i + 1] == self[i] - 1
            ]
        )

    def contract_inc_bonds(self):
        P = Permutation(self)
        while P.num_inc_bonds() > 0:
            for i in range(0, len(P) - 1):
                if P[i + 1] == P[i] + 1:
                    P = Permutation(P[:i] + P[i + 1 :])
                    break
        return P

    def contract_dec_bonds(self):
        P = Permutation(self)
        while P.num_dec_bonds() > 0:
            for i in range(0, len(P) - 1):
                if P[i + 1] == P[i] - 1:
                    P = Permutation(P[:i] + P[i + 1 :])
                    break
        return P

    def contract_bonds(self):
        P = Permutation(self)
        while P.num_bonds() > 0:
            for i in range(0, len(P) - 1):
                if P[i + 1] == P[i] + 1 or P[i + 1] == P[i] - 1:
                    P = Permutation(P[:i] + P[i + 1 :])
                    break
        return P


if __name__ == "__main__":
    pass



class Node(object):
    """A class for nodes of binary plane trees."""

    def __init__(self, label):
        self.label = label
        self.left_child = None
        self.right_child = None

    def __len__(self):
        length = 1
        if self.left_child:
            length += len(self.left_child)
        if self.right_child:
            length += len(self.right_child)
        return length

    def __repr__(self, width=None):
        label = str(self.label)

        if width:
            label = f"{label:{width}s}"
        else:
            width = len(label)

        lines = [label]

        if self.right_child:
            right_repr = self.right_child.__repr__(width=width)
            right_lines = right_repr.split("\n")
            for idx, right_line in enumerate(right_lines):
                try:
                    lines[idx] += " - " + right_line
                except IndexError:
                    lines.append(" " * (width + 3))

        if self.left_child:
            # seen = len(lines)
            left_repr = self.left_child.__repr__(width=width)
            left_lines = left_repr.split("\n")
            for idx, left_line in enumerate(left_lines):
                if idx == 0:
                    lines.append(" " * (width) + r" \ " + left_line)
                else:
                    lines.append(" " * (width + 3) + left_line)

        return "\n".join(lines)

    def width(self):
        raise NotImplementedError

    def depth(self):
        depth_below = 0
        if self.left_child:
            depth_below = max(depth_below, self.left_child.depth())
        if self.right_child:
            depth_below = max(depth_below, self.right_child.depth())
        return depth_below + 1

    def add_left(self, child):
        self.left_child = child

    def add_right(self, child):
        self.right_child = child

    def has_children(self):
        return self.left_child or self.right_child

    def postorder(self):
        reading = []
        if self.left_child:
            reading += self.left_child.postorder()
        if self.right_child:
            reading += self.right_child.postorder()
        reading.append(self.label)
        return reading


def create_tree(p):
    """Given a nonempty permutation `p`, return the associated decreasing binary plane tree."""
    L = list(p)

    max_val = max(p)
    max_idx = p.index(max_val)

    before = L[:max_idx]
    # print(f"before={before}")
    after = L[max_idx + 1 :]
    # print(f"after={after}")

    T = Node(max_val)
    if len(before):
        T.add_left(create_tree(before))
    if len(after):
        T.add_right(create_tree(after))

    return T


if __name__ == "__main__":
    p = [1, 3, 2, 7, 6, 5, 4]
    print(p)
    t = create_tree(p)
    print(t)
    print(t.postorder())

    p = [2, 3, 1]
    print(p)
    t = create_tree(p)
    print(t)
    p = t.postorder()
    print(p)
    t = create_tree(p)
    print(t.postorder())



from .permutation import Permutation
from .permset import PermSet
from .permclass import PermClass
from .propertyclass import PropertyClass
from .avclass import AvClass

# from .pegpermutation import PegPermutation
# from .pegpermset import PegPermSet
from .geometricgridclass import GeometricGridClass

# from .InsertionEncoding import *
# from .RestrictedContainer import *

# try:
#     import matplotlib as mpl

#     mpl.rc("axes", fc="E5E5E5", ec="white", lw="1", grid="True", axisbelow="True")
#     mpl.rc("grid", c="white", ls="-")
#     mpl.rc("figure", fc="white")
#     mpl_imported = True
# except ImportError:
#     print("Install matplotlib for extra plotting functionality.")
#     pass

Perm = Permutation
Av = AvClass



from permpy.permutation import Permutation

# stack


def main1():
    pi = Permutation(3214765)

    iota = Permutation(12345)

    for n in range(len(pi), 10):
        print("n = {}".format(n))
        L = pi.optimizers(n)
        for tau in L:
            if tau.involves(Permutation(312)) or tau.involves(Permutation(231)):
                print("tau = {} is not layered!".format(tau))
            else:
                print([len(layer) for layer in tau.sum_decomposition()])
            if tau.involves(iota):
                print("tau = {} contains {}!".format(tau, iota))
                copies = tau.copies_of(iota)
                for copy in copies:
                    rest_of_tau = [
                        val for idx, val in enumerate(tau) if idx not in copy
                    ]
                    for sigma in L:
                        rest_of_sigma = [
                            val for idx, val in enumerate(sigma) if idx not in copy
                        ]
                        if rest_of_sigma == rest_of_tau:
                            print("tau = {}".format(tau))
                            print("sigma = {}".format(sigma))
                            print("change to = {}".format([sigma[idx] for idx in copy]))


def main2():
    qpi = Permutation(3214765)
    iota = Permutation(12345)

    a = Permutation(312)
    b = Permutation(231)

    height = 10
    L = iota.upset(height, stratified=True)
    for n in range(len(iota), height + 1):
        S = L[n]
        for tau in S:
            if tau.involves(a) or tau.involves(b):
                continue
            num_copies = len(tau.copies_of(pi))
            if num_copies == 0:
                continue
            print("{:30s} {:10d}".format("tau = {}".format(tau), num_copies))
            copies = tau.copies_of(iota)
            for copy in copies:
                print("\tcopy = {}".format(" ".join(str(val + 1) for val in copy)))
                sys.stdout.flush()

                lower_deletion = tau.delete(values=copy[1])
                lower_inflation = lower_deletion.insert(tau.index(copy[0]), copy[0] + 1)
                lower_num_copies = len(lower_inflation.copies_of(pi))

                print(
                    "{:30s} {:6d}".format(
                        "\tnew = {}".format(lower_inflation), lower_num_copies
                    ),
                    end="",
                )

                upper_deletion = tau.delete(values=copy[3])
                upper_inflation = upper_deletion.insert(tau.index(copy[4]) - 1, copy[4])
                upper_num_copies = len(upper_inflation.copies_of(pi))

                print(
                    "{:30s} {:6d}".format(
                        "\tnew = {}".format(upper_inflation), upper_num_copies
                    )
                )
                sys.stdout.flush()

                new_num_copies = max(lower_num_copies, upper_num_copies)

                if 0 < new_num_copies < num_copies:
                    break
            else:
                continue

            print("Something went wrong!")
            sys.stdout.flush()
            break

    def stack_sorted(self):
        """Return self after one deterministic stack sort."""
        before = list(self)
        stack = []  # bottom to top, i.e. stack[0] is the bottom, stack[-1] is the top.
        after = []

        while before:
            val = before.pop(0)

            while stack and val > stack[-1]:
                after.append(stack.pop(-1))

            stack.append(val)

        after += stack[::-1]
        return Permutation(after, clean=True)

    def gen_stack_sorted(self):
        """Yield self after one deterministic stack sort."""
        before = list(self)
        stack = []  # bottom to top, i.e. stack[0] is the bottom, stack[-1] is the top.

        while before:
            val = before.pop(0)

            while stack and val > stack[-1]:
                yield stack.pop(-1)

            stack.append(val)

        for val in stack[::-1]:
            yield val

    # def sort_sequence(self):
    # 	"""Return the sequence obtained by (deterministically) stack sorting self.
    # 	"""
    # 	p = Permutation(self, clean=True)
    # 	L = [self]
    # 	while not p.is_identity():
    # 		p = p.stack_sorted()
    # 		L.append(p)
    # 	return L

    # def create_tree(self):
    # 	"""Given a nonempty permutation `p`, return the associated decreasing binary plane tree.
    # 	"""
    # 	assert self
    # 	L = list(self)

    # 	max_val = max(self)
    # 	max_idx = self.index(max_val)

    # 	before = L[:max_idx]
    # 	after  = L[max_idx+1:]

    # 	T = Node(max_val) #
    # 	if len(before):
    # 		T.add_left(create_tree(before))
    # 	if len(after):
    # 		T.add_right(create_tree(after))

    # 	return T

    def inverse_stack(self):
        L = []
        for pi in Permutation.genall(len(self)):
            for idx, val in enumerate(pi.gen_stack_sorted()):
                if not self[idx] == val:
                    break
            else:
                # For loop didn't break!
                L.append(pi)
        return L

    def stack_preimage(self, k):
        S = permset.PermSet(self)
        for _ in range(k):
            S = S.inverse_stack()
        return S



import types


def av_test(p):
    """Return a function that is True if the input avoids p."""
    return lambda q: p not in q


def copy_func(f, name=None):
    """Return a function with same code, globals, defaults, closure, and
    name (or provide a new name)
    """
    fn = types.FunctionType(
        f.__code__, f.__globals__, name or f.__name__, f.__defaults__, f.__closure__
    )
    # in case f was given attrs (note this dict is a shallow copy):
    fn.__dict__.update(f.__dict__)
    return fn


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['permpy',
 'permpy.InsertionEncoding',
 'permpy.RestrictedContainer',
 'permpy.deprecated',
 'permpy.misc']

package_data = \
{'': ['*']}

install_requires = \
['pytest>=7.1.2,<8.0.0']

setup_kwargs = {
    'name': 'permpy',
    'version': '0.2.0',
    'description': 'A package for analyzing permutation patterns.',
    'long_description': "permpy\n=======\n\n## A Python Permutations Class\n\nContains Various tools for working interactively with permutations. \nEasily extensible.\n\n### Examples:\n```python\n>>> import permpy as pp\n>>> \n>>> p = pp.Perm.random(8)\n>>> p\n 5 4 7 1 6 2 3 8 \n>>> p.cycles()\n'( 6 2 4 1 5 ) ( 7 3 ) ( 8 )'\n>>> p.order()\n10 \n>>> p ** p.order()\n 1 2 3 4 5 6 7 8\n>>>\n>>> S = pp.PermSet.all(6)\n>>> S\nSet of 720 permutations\n>>> S.total_statistic(pp.Perm.num_inversions)\n5400\n>>> S.total_statistic(pp.Perm.num_descents)\n1800\n>>> \n>>> A = pp.AvClass([ 132 ])\n>>> A\n[Set of 0 permutations, \n Set of 1 permutations, \n Set of 2 permutations, \n Set of 5 permutations, \n Set of 14 permutations, \n Set of 42 permutations, \n Set of 132 permutations, \n Set of 429 permutations, \n Set of 1430 permutations]\n>>> \n```\n\n## Build Instructions\nFor a summary of how PermPy is built, go [here](https://py-pkgs.org/03-how-to-package-a-python#summary-and-next-steps).\n```bash\n$ python -m poetry build\n$ python -m poetry publish -r test-pypi\n```",
    'author': 'Michael Engen',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



from collections import Counter, namedtuple

X = namedtuple("X", ["a", "b", "c"])


def next_states(S, mul):
    """Compute the states that would arise from the given state."""
    to_return = Counter()

    def include(a, b, c):
        to_return[X(a, b, c)] += mul

    if S.b != -1:

        if S.c > 0:
            include(S.a, S.b, S.c + 1)

        include(S.b, S.c, 1)

        # for k in range(0,S.a):
        # 	include(k,-1,S.c)

        # for k in range(0,S.b):
        # 	include(k,-1,S.c)
        if S.b > 0:
            for k in range(0, S.a + 1):
                include(k, -1, S.c)

        for k in range(1, S.b):
            include(k, -1, S.c)

        # for k in range(1,S.c):
        # 	include(k, -1, 0)

        for k in range(1, S.c):
            include(0, -1, S.c - k + 1)

    else:

        if S.c > 0:
            include(S.a, -1, S.c + 1)
            include(0, S.c, 1)
        else:
            include(S.a, -1, 1)

        for k in range(0, S.a + 1):
            include(k, -1, S.c)

        for k in range(1, S.c):
            include(k, -1, 0)

    return to_return


def iterate(states):
    """Iterate the current Counter of states once."""
    ns = sum([next_states(S, mul) for S, mul in states.items()], Counter())
    return ns


def states(n):
    """Return the states corresponding to permutations of length n in NS."""
    states = Counter()
    states[X(0, 0, 0)] = 1

    for _ in range(n):
        states = iterate(states)

    return states


def pstring(state, num):
    """Return a nice string give a state and its count.

    Example:
    >>> pstring(X(0,-1,1),4)
    ( 0, *, 1) :  4

    """
    a, b, c = state
    if b == -1:
        b = " *"
    return f"({a:2},{b:2},{c:2}) : {num:2}"


if __name__ == "__main__":
    S = Counter([X(0, 0, 0)])

    for n in range(1, 6):
        print(f"\nn = {n}\n")

        for state, num in sorted(S.items()):
            NS = next_states(state, num)
            if 3 < sum(NS.values()) // num < 5:
                print(f"{pstring(state, num)} -> ")
                for nstate, nnum in sorted(NS.items()):
                    print("\t" + pstring(nstate, nnum // num))
            # print("\t" + pstring(state, num))

        S = iterate(S)

        # if n == 4:
        # 	S[X(0,-1,0)] -= 1

        print(f"{n:2}, {sum(S.values()):10}")



from .vector import *


class VectorSet(list):
    def basis_union(self, B):
        if len(B) > 0 and B[0] == -1:
            return self
        if len(self) > 0 and self[0] == -1:
            return B
        S = set()
        for V in self:
            for W in B:
                S.add(V.meet(W))
        return VectorSet(S).minimal_elements()

    def minimal_elements(self):
        C = self[:]
        for V in self:
            for W in self:
                if V == W:
                    continue
                if W.contained_in(V):
                    C.remove(V)
                    break
        return C

    def meet_all(self):
        if len(self) == 0:
            return Vector([])

        l = len(self[0])
        V = Vector([1] * l)
        for W in self:
            V = V.meet(W)
        return V
